<?php

namespace Lib {
    require_once __DIR__ . '../vendor/autoload.php';

    use \Riverline\MultiPartParser\Part;

    const HOST = 'building.app.lowenergy.jp';
    const PROTOCOL = 'https://';
    const URL_BASE = PROTOCOL . HOST . '/api/v1/';
    const BOUNDARY = "================12345678987654321"; // boundary は適当で良い

    /**
     * POSTするデータを作成します。
     * @param array $files POSTするファイルの配列
     * @return string POSTデータ
     */
    function createPostData(array $files): string
    {
        $postData = '';
        foreach ($files as $file => $data) {
            if (empty($data)) {
                continue;
            }

            $postData .= "\r\n--" . BOUNDARY . "\r\n";
            $postData .= "Content-Disposition: mixed; name=\"$file\"\r\n";
            if ($file == 'request') {
                $postData .= "Content-Type: text/json\r\n";
            } else {
                $postData .= "Content-Type: text/csv\r\n";
            }
            $postData .= "\r\n";
            $postData .= $data;
        }
        $postData .= "--" . BOUNDARY . "--";
        return $postData;
    }

    /**
     * 計算リクエストを送信します。
     * @return string リクエストID
     */
    function requestCompute(): string
    {
        $postFiles = [
            'F0' => getLocalFile('様式0*.csv'),
            'F1' => getLocalFile('様式1*.csv'),
            'F2-1' => getLocalFile('様式2-1*.csv'),
            'F2-2' => getLocalFile('様式2-2*.csv'),
            'F2-3' => getLocalFile('様式2-3*.csv'),
            'F2-4' => getLocalFile('様式2-4*.csv'),
            'F2-5' => getLocalFile('様式2-5*.csv'),
            'F2-6' => getLocalFile('様式2-6*.csv'),
            'F2-7' => getLocalFile('様式2-7*.csv'),
            'F3-1' => getLocalFile('様式3-1*.csv'),
            'F3-2' => getLocalFile('様式3-2*.csv'),
            'F3-3' => getLocalFile('様式3-3*.csv'),
            'F4' => getLocalFile('様式4*.csv'),
            'F5-1' => getLocalFile('様式5-1*.csv'),
            'F5-2' => getLocalFile('様式5-2*.csv'),
            'F6' => getLocalFile('様式6*.csv'),
            'F7-1' => getLocalFile('様式7-1*.csv'),
            'F7-2' => getLocalFile('様式7-2*.csv'),
            'F7-3' => getLocalFile('様式7-3*.csv'),
            'F8' => getLocalFile('様式8*.csv')
        ];

        $postData = createPostData($postFiles);
        $opts = ['http' =>
            [
                'method' => 'POST',
                'header' => implode("\r\n", [
                    'Host: ' . HOST,
                    'Content-Type: multipart/mixed; boundary=' . BOUNDARY,
                    'Content-Length: ' . strlen($postData),
                ]),
                'content' => $postData
            ]
        ];

        $context = stream_context_create($opts);
        echo "[REQUEST] Request to compute...\n";
        $json = file_get_contents(URL_BASE . 'compute', false, $context);
        $result = json_decode($json, true);
        echo "[RESPONSE] Status is '{$result['Status']}'\n";
        echo "\n";
        if ($result['Status'] !== 'OK') {
            echo $json;
            exit(-1);
        }

        return $result['RequestID'];
    }

    /**
     * 計算の実行状況を確認します。
     * @param string $id リクエストID
     * @param bool $pdf エネルギー計算の状況確認ではfalse、PDF生成の状況確認ではtrue
     */
    function checkProgress(string $id, bool $pdf = false)
    {
        $postData = "{\"RequestID\": \"$id\"}";
        $opts = ['http' =>
            [
                'method' => 'POST',
                'header' => implode("\r\n", [
                    'Host: ' . HOST,
                    'Content-Type: text/json',
                    'Content-Length: ' . strlen($postData),
                ]),
                'content' => $postData
            ]
        ];
        $context = stream_context_create($opts);

        do {
            sleep(3);
            echo "[REQUEST] Check " . ($pdf ? "PDF generating" : "computing") . " progress...\n";
            $url = $pdf ? 'pdfGeneratingState' : 'computingState';
            $json = file_get_contents(URL_BASE . $url, false, $context);
            $result = json_decode($json, true);
            echo "[RESPONSE] Status is '{$result['Status']}'\n";
            echo "\n";
        } while ($result['Status'] === 'Computing' || $result['Status'] === 'Generating');

        if ($result['Status'] !== 'Done') {
            echo 'Caused an error when checking progress.';
            exit(-1);
        }
    }

    /**
     * 結果を取得します。
     * @param string $id リクエストID
     * @param bool $pdf 計算結果の取得ではfalse、PDFの取得ではtrue
     */
    function fetchResult(string $id, bool $pdf = false)
    {
        $postData = "{\"RequestID\": \"$id\"}";
        $opts = ['http' =>
            [
                'method' => 'POST',
                'header' => implode("\r\n", [
                    'Host: ' . HOST,
                    'Content-Type: text/json',
                    'Content-Length: ' . strlen($postData),
                ]),
                'content' => $postData
            ]
        ];
        $context = stream_context_create($opts);
        echo "[REQUEST] Fetch results...\n";
        $url = $pdf ? 'pdfDownload' : 'result';
        $result = file_get_contents(URL_BASE . $url, false, $context);

        if ($result === false) {
            echo "[RESPONSE] Error. Response headers are:\n";
            var_dump($http_response_header);
            echo "\n";
            exit(-1);
        }
        return implode("\r\n", $http_response_header) . "\r\n\r\n" . $result;
    }

    /**
     * Globを使ってローカルファイルの内容を取得します。
     *
     * @param string $glob ローカルファイルを検索するためのグロブ
     * @return string? ローカルファイルの内容。ただし、ファイルが見つからなかった場合はnullを返します。
     */
    function getLocalFile(string $glob): ?string
    {
        $files = glob($glob);
        if (empty($files)) {
            return null;
        }
        return file_get_contents(solveShiftJISFilePath($files[0]));
    }

    /**
     * 計算結果のCSVとJsonを保存します。
     *
     * @param string $response APIの応答
     */
    function saveResult(string $response): void
    {
        $contents = new Part($response);
        if ($contents->isMultiPart()) {
            foreach ($contents->getParts() as $part) {
                $contentType = strtolower($part->getHeader('Content-Type'));
                $filename = $part->getName();
                switch ($contentType) {
                    case 'text/csv':
                        $filename .= '.csv';
                        break;
                    case 'text/json':
                        $filename .= '.json';
                        break;
                }
                echo "Save '$filename'\n";
                file_put_contents($filename, $part->getBody());
            }
        }
        echo "\n";
    }

    /**
     * PDFファイルを保存します。
     *
     * @param string $response APIの応答
     */
    function savePdf(string $response): void
    {
        try {
            $contents = new Part($response);
        } catch (\InvalidArgumentException $e) {
            echo "Error.\n";
            echo "Response is:\n";
            var_dump($response);
            exit(-1);
        }
        if ($contents->isMultiPart()) {
            foreach ($contents->getParts() as $part) {
                $filename = $part->getName();
                echo "Save '$filename'\n";
                file_put_contents($filename, $part->getBody());
            }
        }
    }

    /**
     * Windowsの日本語ファイルを正しく扱えるように、ファイルパスを変換します。
     * まずフォルダ区切り文字を'\'から'/'に変換します。
     * その後、0x5cを検出し、直後にエスケープ文字'\'を追加します。
     *
     * @param string $filePath ファイルパス
     * @return string Windowsで日本語ファイルも扱えるファイルパス
     */
    function solveShiftJISFilePath(string $filePath): string
    {
        // Windowsでなければ何もしない。
        if (!defined('PHP_WINDOWS_VERSION_MAJOR')) {
            return $filePath;
        }

        $filePathSeperatedBySlashes = str_replace("\\", '/', $filePath);
        $bytes = unpack('C*', $filePathSeperatedBySlashes);
        for ($i = 1; $i <= count($bytes); $i++) {
            if ($bytes[$i] === 0x5c) {
                array_splice($bytes, $i + 1, 0, [0x5c]);
                $i++;
            }
        }
        $solvedFilePath = implode(array_map("chr", $bytes));
        return $solvedFilePath;
    }

    function main(): void
    {
        $requestId = requestCompute();

        // Check and fetch for computing
        checkProgress($requestId);
        saveResult(fetchResult($requestId));

        // Check and fetch for PDF
        checkProgress($requestId, true);
        savePdf(fetchResult($requestId, true));
    }
}

namespace {
    Lib\main();
}
